"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
from datetime import datetime
from rfc822 import mktime_tz, parsedate_tz

import isodate
import shortuuid
from pyld import jsonld
from rdflib import URIRef, XSD, RDFS, Graph
from rdflib.term import Literal, BNode

__author__ = 'Fernando Serena'


def ld_triples(ld, g=None):
    bid_map = {}

    def parse_term(term):
        if term['type'] == 'IRI':
            return URIRef(term['value'])
        elif term['type'] == 'literal':
            datatype = URIRef(term.get('datatype', None))
            if datatype == XSD.dateTime:
                try:
                    term['value'] = float(term['value'])
                    term['value'] = datetime.utcfromtimestamp(term['value'])
                except:
                    try:
                        term['value'] = isodate.parse_datetime(term['value'])
                    except:
                        timestamp = mktime_tz(parsedate_tz(term['value']))
                        term['value'] = datetime.fromtimestamp(timestamp)
            if datatype == RDFS.Literal:
                datatype = None
                try:
                    term['value'] = float(term['value'])
                except:
                    pass
            return Literal(term['value'], datatype=datatype)
        else:
            bid = term['value'].split(':')[1]
            if bid not in bid_map:
                bid_map[bid] = shortuuid.uuid()
            return BNode(bid_map[bid])

    if g is None:
        g = Graph()
    norm = jsonld.normalize(ld)
    def_graph = norm.get('@default', [])
    for triple in def_graph:
        subject = parse_term(triple['subject'])
        predicate = parse_term(triple['predicate'])
        if not predicate.startswith('http'):
            continue
        object = parse_term(triple['object'])
        g.add((subject, predicate, object))

    return g



import json
import urlparse
from urllib import quote, unquote
import shelve
import os
import requests
import sys
from SPARQLWrapper import JSON
from SPARQLWrapper import SPARQLWrapper
from rdflib import Graph, Namespace
from concurrent.futures import ThreadPoolExecutor, wait

from kg_search.ld import ld_triples
from kg_search import kg_cache, wd_cache, app

SCHEMA = Namespace('http://schema.org/')
GOOGLE_API_KEY = os.environ.get('GOOGLE_API_KEY')
if not GOOGLE_API_KEY:
    sys.exit(-1)

wiki_d = shelve.open('wiki-entities')

pool = ThreadPoolExecutor()


@wd_cache.memoize(3600)
def search_wiki_entity(wiki):
    sparql = SPARQLWrapper("https://query.wikidata.org/sparql")
    sparql.setReturnFormat(JSON)

    wiki = str(wiki)
    # if wiki not in wiki_d:
    sparql.setQuery("""
       prefix schema: <http://schema.org/>
       SELECT * WHERE {
           <%s> schema:about ?item .
       }
   """ % wiki)

    entity = None
    try:
        results = sparql.query().convert()

        for result in results["results"]["bindings"]:
            entity = result["item"]["value"].replace('http://www.wikidata.org/entity/', '')
            if entity:
                print 'found {} for {}'.format(entity, wiki)
            break
    except Exception:
        pass

        # wiki_d[wiki] = entity

    return entity
    # return wiki_d[wiki]


def search_dbpedia_uri(wiki):
    wiki = str(wiki)
    parse = urlparse.urlparse(wiki)
    dbpedia_path = parse.path.replace('wiki', 'resource')
    return 'http://dbpedia.org' + dbpedia_path


def search_types_in_dbpedia(dbpedia_uri):
    sparql = SPARQLWrapper("http://dbpedia.org/sparql")
    sparql.setReturnFormat(JSON)

    sparql.setQuery("""           
           SELECT * WHERE {
               <%s> rdf:type ?type .
           }
       """ % dbpedia_uri)

    try:
        results = sparql.query().convert()

        for result in results["results"]["bindings"]:
            ty = result["type"]["value"]
            if ty.startswith('http://schema.org/'):
                yield str(ty).replace('http://schema.org/', '')
    except Exception:
        pass


def enrich_wiki_entry(wiki, name, types):
    entity = search_wiki_entity(wiki)
    dbpedia = search_dbpedia_uri(wiki)
    enrich_types = types
    if len(types) == 1 and 'Thing' in types:
        dbpedia_types = set(search_types_in_dbpedia(dbpedia))
        enrich_types = dbpedia_types.union(set(types))

    return enrich_types, entity, dbpedia


def iriToUri(iri):
    parts = urlparse.urlparse(iri)
    return urlparse.urlunparse(
        part.encode('utf-8') if parti == 1 else quote(part.encode('utf-8'))
        for parti, part in enumerate(parts)
    )


@kg_cache.memoize(3600)
def kg_search(q, types=None, count=None):
    kg_request_url = 'https://kgsearch.googleapis.com/v1/entities:search?query={}&key={}&indent=True&limit={}'.format(
        q, GOOGLE_API_KEY, count)

    if isinstance(types, list):
        types = ','.join(types)
        kg_request_url += '&types={}'.format(types)

    kg_response = requests.get(kg_request_url)

    kg = Graph()
    kg.bind('schema', SCHEMA)
    ld_triples(kg_response.json(), kg)

    print kg.serialize(format='turtle')

    kg_query_result = kg.query("""
                SELECT DISTINCT (MIN(?score) as ?min) (MAX(?score) as ?max) (AVG(?score) as ?avg) WHERE {
                   [] <http://schema.googleapis.com/resultScore> ?score
                }
            """)

    try:
        max_min_score = list(kg_query_result).pop()
        max_score, min_score, avg_score = max_min_score.max.toPython(), \
                                          max_min_score.min.toPython(), \
                                          max_min_score.avg.toPython()

    except:
        max_score, min_score, avg_score = 0, 0, 0

    score_th = avg_score * 0.1 + min_score * 0.9
    print max_score, min_score, avg_score, score_th

    kg_query_result = kg.query("""
            SELECT DISTINCT ?wiki ?type ?score ?name WHERE {
               [] <http://schema.org/result> [
                     a ?type ;
                     <http://schema.googleapis.com/detailedDescription> [
                        <http://schema.org/url> ?wiki
                     ] ;
                     <http://schema.org/name> ?name
                  ] ;
                  <http://schema.googleapis.com/resultScore> ?score
            }
        """)

    res_dict = {}
    for kgr in kg_query_result:
        if kgr.score.toPython() >= score_th:
            print kgr.score, kgr.wiki
            wiki_uri = iriToUri(kgr.wiki)
            wiki_uri = unquote(wiki_uri).decode('utf8')
            if (wiki_uri, kgr.name) not in res_dict:
                res_dict[(wiki_uri, kgr.name)] = set([])
            res_dict[(wiki_uri, kgr.name)].add(kg.qname(kgr.type).split(':')[1])

    return res_dict


def search_entities(q, **kwargs):
    kg_results = kg_search(q, **kwargs)

    results = []
    futures = []

    for (wiki, name), types in kg_results.items():
        future = pool.submit(enrich_wiki_entry, wiki=wiki, name=name, types=types)
        futures.append(future)
        results.append((future, wiki, name))

    with app.app_context():
        wait(futures)
        for future, wiki, name in results:
            types, entity, dbpedia = future.result()
            yield (types, entity, dbpedia, wiki, name)


@kg_cache.memoize(3600)
def search_seeds_from_image(img, types=None):
    r = requests.post(
        'https://vision.googleapis.com/v1/images:annotate?key={}'.format(GOOGLE_API_KEY),
        data=json.dumps({
            "requests": [
                {
                    "image": {
                        "source": {
                            "imageUri": img
                        }
                    },
                    "features": [
                        {
                            "type": "WEB_DETECTION"
                        }
                    ]
                }
            ]
        }))

    if r.status_code == 200:
        data = r.json()
        try:
            descriptions = map(lambda x: x['description'], data['responses'][0]['webDetection']['webEntities'])
            for d in descriptions:
                for seed_tuple in search_seeds(d, types=types, count=1):
                    yield seed_tuple
        except:
            pass


def search_seeds(search, types=None, count=200):
    if not types:
        for entity_types, q, db, wiki, name in search_entities(search, count=count):
            yield entity_types, q, db, wiki, name
    else:
        for type in types:
            for entity_types, q, db, wiki, name in search_entities(search, types=[type], count=count):
                yield entity_types, q, db, wiki, name



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import traceback

from flask import request
from flask.json import jsonify

from kg_search.search import search_seeds_from_image, search_seeds
from kg_search import app, cache

__author__ = 'Fernando Serena'


def make_cache_key(*args, **kwargs):
    path = request.path
    qargs = dict(request.args.items())
    args = ''.join(['{}{}'.format(k, qargs[k]) for k in sorted(qargs.keys())])
    return (path + args).encode('utf-8')


@app.route('/search')
@cache.cached(timeout=3600, key_prefix=make_cache_key)
def search():
    try:
        q = request.args.get('q')
        img = request.args.get('img')
        types = request.args.getlist('types')
        limit = request.args.get('limit', 200)
        entities = {}
        if img is not None:
            gen = search_seeds_from_image(img, types=types)
        else:
            gen = search_seeds(q, types=types, count=limit)
        for types, q, db, wiki, name in gen:
            for t in types:
                if t not in entities:
                    entities[t] = []
                entities[t].append({'wikidata': q, 'name': name, 'dbpedia': db, 'wikipedia': wiki})
        return jsonify(entities)
    except Exception:
        traceback.print_exc()


if __name__ == '__main__':
    app.run(host='0.0.0.0', port=5015, use_reloader=False, debug=False, threaded=True)
